#include <hmg.ch>
#include <GraphPlus.ch>

#include "hbclass.ch"

CLASS GraphPlus

   DATA aData INIT {} // Graph data
   DATA aLegends INIT {}
   DATA aCategories INIT {} // Horizontal Axis
   DATA nGraphType INIT 1 // Columns
   DATA cTitle INIT '' // Graph Title
   DATA cXTitle INIT '' // Horizontal Title
   DATA cYTitle INIT '' // Vertical Title
   DATA lXRotate INIT .F. // xValues Rotate
   DATA nHeight INIT 600 // Graph Height
   DATA nWidth INIT 600 // Graph Width
   DATA lShowLegends INIT .F. // Show legends
   DATA lDataLabels INIT .F. // show data labels
   DATA lHGrid INIT .T. // Horizontal Grid
   DATA lVGrid INIT .T. // Vertical Grid
   DATA lShowValues INIT .T. // Show Value Axis
   DATA lShowCategories INIT .T. // show categories
   DATA nLegendPos INIT 1 // 0 Right 1 Bottom
   DATA lOutBorder INIT .T. // draw outside border
   DATA aColors INIT {} // aColors
   DATA aTitleFont INIT { 'Arial', 12, .T., .F., .F., .F. }
   DATA aLegendFont INIT { 'Arial', 8, .F., .F., .F., .F. }
   DATA aBackColor INIT { 255, 255, 255 }
   DATA aTitleColor INIT { 0, 0, 255 }
   DATA aTextColor INIT { 0, 0, 0 }
   DATA aGridLineColor INIT { 200, 200, 200 }
   DATA nLegendBoxSize INIT 10
   DATA nLegendWidth INIT 0
   DATA nColorTheme INIT 1
   DATA nUpper INIT 0
   DATA nLower INIT 0
   DATA lCategoryBelow INIT .T.
   DATA nRange INIT 0
   DATA nPixelperPoint INIT 0
   DATA nLineWidth INIT 2
   DATA nZeroPos INIT 0
   DATA nVDivision INIT 5
   DATA nHDivision INIT 0
   DATA nUpperDivision INIT 0
   DATA nLowerDivision INIT 0
   DATA cPicture INIT '99999'
   DATA nCategoryRow INIT 0
   DATA nBarGapRatio INIT 0.2 // in percentage 20%
   DATA nGapWidthRatio INIT 1 // in percentage 200%
   DATA nImageHeight INIT 600
   DATA nImageWidth INIT 600
   DATA nPointSize INIT 10

   DATA hBitMap INIT NIL // Bitmap to store Graph
   DATA hDC INIT NIL
   DATA BTStruct INIT NIL
   DATA nTop INIT 10
   DATA nLeft INIT 10
   DATA nRight INIT 590
   DATA nBottom INIT 590
   DATA aScale INIT {}
   DATA aColorThemes INIT { { CLR_HAVELOCK_BLUE, CLR_WESTSIDE, CLR_LIGHT_GREY, CLR_DARK_YELLOW, CLR_PICTON_BLUE, CLR_APPLE }, ;
      { CLR_HAVELOCK_BLUE, CLR_LIGHT_GREY, CLR_PICTON_BLUE, CLR_DARKER_BLUE, CLR_DARK_GREY, CLR_ENDEAVOUR }, ;
      { CLR_WESTSIDE, CLR_DARK_YELLOW, CLR_APPLE, CLR_SADDLE_BROWN, CLR_LIGHT_BROWN, CLR_DELL }, ;
      { CLR_APPLE, CLR_PICTON_BLUE, CLR_DARK_YELLOW, CLR_DELL, CLR_ENDEAVOUR, CLR_LIGHT_BROWN }, ;
      { CLR_HAVELOCK_BLUE }, ;
      { CLR_WESTSIDE }, ;
      }
   DATA lLighter INIT .T.

   METHOD New
   METHOD Draw
   METHOD Bitmap
   METHOD DrawColumnGraph
   METHOD DrawBarGraph
   METHOD DrawLineGraph
   METHOD DrawPieGraph
   METHOD DrawPointsGraph
   METHOD DrawLegends
   METHOD InitColors
   METHOD DrawXYTitles
   METHOD DrawCategories
   METHOD DrawVerticalScale
   METHOD DrawHorizontalGrid
   METHOD DrawVerticalGrid
   METHOD SwitchRowsToColumns
   METHOD Save
   METHOD GraphData SETGET
   METHOD GridData SETGET
   METHOD Legends SETGET
   METHOD Picture SETGET
   METHOD Categories SETGET
   METHOD GraphColors SETGET
   METHOD GraphType SETGET
   METHOD Title SETGET
   METHOD XTitle SETGET
   METHOD YTitle SETGET
   METHOD Height SETGET
   METHOD Width SETGET
   METHOD ShowLegends SETGET
   METHOD ShowLabels SETGET
   METHOD ShowHGrid SETGET
   METHOD ShowVGrid SETGET
   METHOD ShowValues SETGET
   METHOD LegendPos SETGET
   METHOD ColorTheme SETGET
   METHOD BackColor SETGET

END CLASS

METHOD New CLASS GraphPlus
RETURN self

METHOD GraphData( aData ) CLASS GraphPlus

   IF PCount() == 0
      RETURN ::aData
   ELSE
      IF ValType( aData[ 1 ] ) == 'A' // 2D array
         ::aData := aData
      ELSE
         ::aData := { aData } // 1D - make it 2D
      ENDIF
   ENDIF

RETURN Self

METHOD GridData( cFormName, cControlName ) CLASS GraphPlus

   LOCAL i
   LOCAL j
   LOCAL nRowCount := GetProperty( cFormName, cControlName, 'ITEMCOUNT' )
   LOCAL nColumnCount
   LOCAL aLineData
   LOCAL aLine := {}
   IF PCount() == 0
      RETURN ::aData
   ELSE
      ASize( ::aData, 0 )
      ASize( ::aLegends, 0 )
      ASize( ::aCategories, 0 )
      nColumnCount := GetProperty( cFormName, cControlName, 'COLUMNCOUNT' )
      FOR i := 1 TO nColumnCount
         IF i > 1
            AAdd( ::aCategories, GetProperty( cFormName, cControlName, 'COLUMNHEADER', i ) )
         ENDIF
      NEXT i
      IF nRowCount > 0
         FOR i := 1 TO nRowCount
            aLineData := GetProperty( cFormName, cControlName, 'ITEM', i )
            ASize( aLine, 0 )
            FOR j := 1 TO nColumnCount
               IF j == 1
                  AAdd( ::aLegends, aLineData[ j ] )
               ELSE
                  AAdd( aLine, Val( AllTrim( aLineData[ j ] ) ) )
               ENDIF
            NEXT j
            AAdd( ::aData, AClone( aLine ) )
         NEXT i
      ENDIF
   ENDIF

RETURN self

METHOD Bitmap() CLASS GraphPlus
RETURN ::hBitMap

METHOD GraphColors( aColors ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::aColors
   ELSE
      ::aColors := aColors
   ENDIF

RETURN Self

METHOD Legends( aLegends ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::aLegends
   ELSE
      ::aLegends := aLegends
   ENDIF

RETURN Self

METHOD Categories( aCategories ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::aCategories
   ELSE
      ::aCategories := aCategories
   ENDIF

RETURN Self

METHOD BackColor( aBackColor ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::aBackColor
   ELSE
      ::aBackColor := aBackColor
   ENDIF

RETURN Self

METHOD GraphType( nGraphType ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::nGraphType
   ELSE
      ::nGraphType := nGraphType
   ENDIF

RETURN Self

METHOD Title( cTitle ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::cTitle
   ELSE
      ::cTitle := cTitle
   ENDIF

RETURN Self

METHOD XTitle( cXTitle ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::cXTitle
   ELSE
      ::cXTitle := cXTitle
   ENDIF

RETURN Self

METHOD YTitle( cYTitle ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::cYTitle
   ELSE
      ::cYTitle := cYTitle
   ENDIF

RETURN Self

METHOD Height( nHeight ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::nImageHeight
   ELSE
      ::nImageHeight := nHeight
   ENDIF

RETURN Self

METHOD Width( nWidth ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::nImageWidth
   ELSE
      ::nImageWidth := nWidth
   ENDIF

RETURN Self

METHOD ShowLegends( lShowLegends ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::lShowLegends
   ELSE
      ::lShowLegends := lShowLegends
   ENDIF

RETURN Self

METHOD ShowLabels( lDataLabels ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::lDataLabels
   ELSE
      ::lDataLabels := lDataLabels
   ENDIF

RETURN Self

METHOD ShowValues( lShowValues ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::lShowValues
   ELSE
      ::lShowValues := lShowValues
   ENDIF

RETURN Self

METHOD ShowHGrid( lHGrid ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::lHGrid
   ELSE
      ::lHGrid := lHGrid
   ENDIF

RETURN Self

METHOD ShowVGrid( lVGrid ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::lVGrid
   ELSE
      ::lVGrid := lVGrid
   ENDIF

RETURN Self

METHOD Picture( cPicture ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::cPicture
   ELSE
      ::cPicture := cPicture
   ENDIF

RETURN Self


METHOD LegendPos( nLegendPos ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::nLegendPos
   ELSE
      ::nLegendPos := nLegendPos
   ENDIF

RETURN Self

METHOD ColorTheme( nColorTheme ) CLASS GraphPlus
   IF PCount() == 0
      RETURN ::nColorTheme
   ELSE
      ::nColorTheme := nColorTheme
   ENDIF

RETURN Self


METHOD Draw CLASS GraphPlus

   LOCAL aSize, nLowerUpperRatio
   LOCAL nDistance
   LOCAL nRow, nPixels
   LOCAL nCount
   LOCAL nMax, nMin, nI, nJ
   LOCAL nDivision, cDivision, nDecimalLength, nDecimal, cDecimal
   LOCAL nMaxHeight, i
   ::nTop := 10
   ::nLeft := 10
   ::nHeight := ::nImageHeight
   ::nWidth := ::nImageWidth
   ::nBottom := ::nHeight - 10
   ::nRight := ::nWidth - 10
   // init bitmap
   ::hBitmap := BT_BitmapCreateNew ( ::nWidth, ::nHeight, ::aBackColor )
   ::hDC := BT_CreateDC( ::hBitmap, BT_HDC_BITMAP, @::BTStruct )

   ::aTextColor := GetNegativeColor ( ::aBackColor )

   // draw outside border
   IF ::lOutBorder
      BT_DrawRectangle ( ::hDC, 0, 0, ::nWidth - 1, ::nHeight - 1, ::aTextColor, 1 )
   ENDIF
   ::InitColors() // init colors
   // draw legends
   IF ::lShowLegends .AND. hmg_len( ::aLegends ) > 0
      ::DrawLegends()
   ENDIF
   // draw title
   IF hmg_len( ::cTitle ) > 0
      aSize := BT_DrawTextSize ( ::hDC, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ) )
      BT_DrawText ( ::hDC, ::nTop / 2, ( ::nWidth - if( ::lShowLegends .AND. ::nLegendPos < 1, ::nLegendWidth, 0 ) ) / 2, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
      ::nTop := ::nTop + aSize[ 2 ]
   ELSE
      ::nTop := ::nTop
   ENDIF

   IF ::nGraphType <> 4 // not pie
      ::DrawXYTitles() // draw x/y titles
      // draw graph border
      // BT_DrawRectangle (::hDC, ::nTop, ::nLeft, ::nRight - ::nLeft , ::nBottom - ::nTop, ::aTextColor, 1 )

      // find max value
      nMax := 0
      FOR nJ := 1 TO HMG_LEN( ::aData )
         FOR nI := 1 TO HMG_LEN( ::aData[ nJ ] )
            nMax := Max( ::aData[ nJ ][ nI ], nMax )
         NEXT nI
      NEXT nJ
      // find min value
      nMin := 0
      FOR nJ := 1 TO HMG_LEN( ::aData )
         FOR nI := 1 TO HMG_LEN( ::aData[ nJ ] )
            nMin := Min( ::aData[ nJ ][ nI ], nMin )
         NEXT nI
      NEXT nJ
      IF nMax > 0
         ::nUpper := nMax
      ELSE
         ::nUpper := 0
      ENDIF
      IF nMin < 0
         ::nLower := nMin
      ELSE
         ::nLower := 0
      ENDIF
      ::nRange := ::nUpper - ::nLower
      IF ::nUpper <> 0 .AND. ::nLower <> 0
         IF Abs( ::nLower ) < ::nUpper
            nLowerUpperRatio := Abs( ::nLower ) / ::nRange
            ::nLowerDivision := ::nVDivision * nLowerUpperRatio
            IF ::nLowerDivision - Int( ::nLowerDivision ) > 0.0
               ::nLowerDivision := Int( ::nLowerDivision ) + 1
               ::nUpperDivision := ::nVDivision - ::nLowerDivision
            ENDIF
            nDivision := ::nUpper / ::nUpperDivision
         ELSE
            nLowerUpperRatio := Abs( ::nUpper ) / ::nRange
            ::nUpperDivision := ::nVDivision * nLowerUpperRatio
            IF ::nUpperDivision - Int( ::nUpperDivision ) > 0.0
               ::nUpperDivision := Int( ::nUpperDivision ) + 1
               ::nLowerDivision := ::nVDivision - ::nUpperDivision
            ENDIF
            nDivision := Abs( ::nLower ) / ::nLowerDivision
         ENDIF
      ELSE
         nDivision := ::nRange / ::nVDivision
         IF ::nUpper > 0
            ::nUpperDivision := ::nVDivision
            ::nLowerDivision := 0
         ELSE
            ::nLowerDivision := ::nVDivision
            ::nUpperDivision := 0
         ENDIF
      ENDIF
      IF Int( nDivision ) > 0
         cDivision := AllTrim( Str( Int( nDivision ), 14, 0 ) )
         IF hmg_len( cDivision ) > 1 // if more than 1 digits round to power of 10s
            nDivision := ( Val( Left( cDivision, 2 ) ) + 1 ) * 10 ^ ( hmg_len( cDivision ) - 2 )
         ELSE
            IF nDivision > Int( nDivision )
               nDivision++
            ENDIF
         ENDIF
      ELSE //
         cDivision := Transform( nDivision, '9.9999999' )
         nDecimalLength := 8 - hmg_len( AllTrim( Str( Val( Right( cDivision, 7 ) ) ) ) )
         nDecimal := Val( Right( cDivision, 7 ) )
         cDecimal := hb_ntos( nDecimal )
         nDivision := ( Val( AllTrim( Left( cDecimal, 1 ) ) ) + 1 ) / ( 10 ^ nDecimalLength )
      ENDIF
      ::nRange := ::nVDivision * nDivision
      ::nTop := ::nTop + 20
      ::nBottom := ::nBottom - 20
      ::nRight := ::nRight - 5
      IF ::lShowCategories
         nMaxHeight := 0
         FOR i := 1 TO hmg_len( ::aCategories )
            aSize := BT_DrawTextSize ( ::hDC, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ) )
            nMaxHeight := Max( nMaxHeight, aSize[ 2 ] )
         NEXT i
         ::nCategoryRow := ::nBottom
         ::nBottom := ::nBottom - nMaxHeight
      ELSE
         ::nBottom := ::nBottom - 20
      ENDIF
      ::nLeft := ::nLeft + 10
      ::nPixelperPoint := ( ( ::nBottom - ::nTop ) / ::nRange )
      // find zero position
      IF ::nLower == 0 // all values are in positive
         IF ::lShowCategories .AND. hmg_len( ::aCategories ) > 0
            ::lCategoryBelow := .T.
            ::nZeroPos := ::nBottom
         ENDIF
      ELSEIF ::nUpper == 0 // all values are in negative
         IF ::lShowCategories .AND. hmg_len( ::aCategories ) > 0
            ::lCategoryBelow := .F.
            ::nZeroPos := ::nTop
         ENDIF
      ELSE // some values +ve and some -ve
         ::nZeroPos := ::nTop + ( ::nPixelperPoint * ::nRange / ::nVDivision * ::nUpperDivision )
      ENDIF
      ASize( ::aScale, 0 )
      nDistance := ::nRange / ::nVDivision
      nRow := ::nTop
      nPixels := ( ::nBottom - ::nTop ) / ::nVDivision
      FOR i := ::nUpperDivision TO 1 STEP -1
         AAdd( ::aScale, { AllTrim( Transform( nDistance * i, ::cPicture ) ), nRow } )
         nRow += nPixels
      NEXT i
      AAdd( ::aScale, { AllTrim( Transform( 0, ::cPicture ) ), nRow } )
      nRow += nPixels
      nCount := 1
      FOR i := ::nLowerDivision TO 1 STEP -1
         AAdd( ::aScale, { '-' + AllTrim( Transform( nDistance * nCount, ::cPicture ) ), nRow } )
         nCount++
         nRow += nPixels
      NEXT i
      // vertical scale
      IF ::lShowValues
         ::DrawVerticalScale()
      ENDIF
      IF ::lHGrid
         ::DrawHorizontalGrid()
      ENDIF
      IF ::lShowCategories
         ::DrawCategories()
      ENDIF
      IF ::lVGrid
         ::DrawVerticalGrid()
      ENDIF
   ENDIF
   DO CASE
   CASE ::nGraphType == 1 // Columns
      ::DrawColumnGraph()
   CASE ::nGraphType == 2 // Line
      ::DrawLineGraph()
   CASE ::nGraphType == 3 // Points
      ::DrawPointsGraph()
   CASE ::nGraphType == 4 // Pie
      ::DrawPieGraph()
   CASE ::nGraphType == 5 // Bar
      ::DrawBarGraph()
   ENDCASE
   BT_DeleteDC( ::BTstruct )

RETURN NIL

METHOD DrawColumnGraph CLASS GraphPlus

   LOCAL nCategories := 0
   LOCAL nRow, nCol, nData
   LOCAL nTotalBars
   LOCAL nAvailableWidth
   LOCAL nTotalBarRatio := 0
   LOCAL nCategoryRatio

   LOCAL nBarWidth
   LOCAL nCategoryWidth
   LOCAL nBarGap
   LOCAL i, j, nBarHeight

   IF hmg_len( ::aData ) > 0
      IF ValType( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      ENDIF
   ENDIF
   IF nCategories == 0
      RETURN NIL
   ENDIF
   nTotalBars := nCategories * hmg_len( ::aData )
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nTotalBars * 1 // 100%
   nTotalBarRatio := nTotalBarRatio + ( nTotalBarRatio - nCategories ) * ::nBarGapRatio
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth * ::nGapWidthRatio
   nBarGap := nBarWidth * ::nBarGapRatio

   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   FOR i := 1 TO hmg_len( ::aData[ 1 ] )
      FOR j := 1 TO hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarHeight := Abs( nData ) * ::nPixelperPoint
         IF nData > 0
            // positive
            BT_DrawFillRectangle ( ::hDC, nRow - nBarHeight, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            IF ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15, nCol + ( nBarWidth / 2 ), AllTrim( Transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            ENDIF
         ELSE
            // negative
            BT_DrawFillRectangle ( ::hDC, nRow, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            IF ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3, nCol + ( nBarWidth / 2 ), if( nData < 0, '-', '' ) + AllTrim( Transform( Abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            ENDIF
         ENDIF
         nCol += nBarWidth
         IF j < hmg_len( ::aData )
            nCol += nBarGap
         ENDIF
      NEXT j
      nCol += nCategoryWidth
   NEXT i

RETURN NIL

METHOD DrawVerticalScale CLASS GraphPlus

   LOCAL aSize, i
   LOCAL nMaxWidth := 0

   FOR i := 1 TO hmg_len( ::aScale )
      aSize := BT_DrawTextSize ( ::hDC, ::aScale[ i, 1 ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ) )
      nMaxWidth := Max( nMaxWidth, aSize[ 1 ] )
   NEXT i
   ::nLeft := ::nLeft + nMaxWidth
   FOR i := 1 TO hmg_len( ::aScale )
      BT_DrawText ( ::hDC, ::aScale[ i, 2 ] + 2, ::nLeft, ::aScale[ i, 1 ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_RIGHT )
   NEXT i
   ::nLeft := ::nLeft + 5
   // BT_DrawLine (::hDC, ::nTop, nCol + 5, ::nBottom, nCol + 5, ::aTextColor, ::nLineWidth)

RETURN NIL

METHOD DrawHorizontalGrid CLASS GraphPlus

   LOCAL i
   FOR i := 1 TO hmg_len( ::aScale )
      BT_DrawLine ( ::hDC, ::aScale[ i, 2 ], ::nLeft, ::aScale[ i, 2 ], ::nRight, ::aGridLineColor, ::nLineWidth )
   NEXT i

RETURN NIL

METHOD DrawVerticalGrid CLASS GraphPlus

   LOCAL nCategories
   LOCAL nCol
   LOCAL nAvailableWidth
   LOCAL nTotalBarRatio
   LOCAL nBarWidth
   LOCAL nCategoryWidth
   LOCAL i

   nCategories := hmg_len( ::aData[ 1 ] )
   IF nCategories == 0
      RETURN NIL
   ENDIF
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories
   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth
   BT_DrawLine ( ::hDC, ::nTop, ::nLeft, ::nBottom, ::nLeft, ::aGridLineColor, ::nLineWidth )
   nCol := ::nLeft
   FOR i := 1 TO nCategories
      nCol += nCategoryWidth
      BT_DrawLine ( ::hDC, ::nTop, nCol, ::nBottom, nCol, ::aGridLineColor, ::nLineWidth )
   NEXT i

RETURN NIL


METHOD DrawLineGraph CLASS GraphPlus

   LOCAL nCategories := 0
   LOCAL nRow, nCol, nData, nFirstCol, nPrevRow := 0, nPrevCol := 0
   LOCAL nAvailableWidth
   LOCAL nTotalBarRatio

   LOCAL nBarWidth, nBarHeight
   LOCAL nCategoryWidth
   LOCAL i, j

   IF hmg_len( ::aData ) > 0
      IF ValType( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      ENDIF
   ENDIF
   IF nCategories == 0
      RETURN NIL
   ENDIF
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth
   nFirstCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   FOR j := 1 TO hmg_len( ::aData )
      nCol := nFirstCol
      FOR i := 1 TO hmg_len( ::aData[ 1 ] )
         nData := ::aData[ j, i ]
         nBarHeight := Abs( nData ) * ::nPixelperPoint
         IF nData > 0
            // positive
            BT_DrawFillEllipse ( ::hDC, nRow - nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            IF ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2, nCol, AllTrim( Transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            ENDIF
            IF i <> 1 // first point
               BT_DrawLine ( ::hDC, nPrevRow, nPrevCol, nRow - nBarHeight, nCol, ::aColors[ j ], ::nLineWidth )
            ENDIF
            nPrevRow := nRow - nBarHeight
            nPrevCol := nCol
         ELSE
            // negative
            BT_DrawFillEllipse ( ::hDC, nRow + nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            IF ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2, nCol, if( nData < 0, '-', '' ) + AllTrim( Transform( Abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            ENDIF
            IF i <> 1
               BT_DrawLine ( ::hDC, nPrevRow, nPrevCol, nRow + nBarHeight, nCol, ::aColors[ j ], ::nLineWidth )
            ENDIF
            nPrevRow := nRow + nBarHeight
            nPrevCol := nCol
         ENDIF
         nCol += nCategoryWidth
      NEXT i
   NEXT j

RETURN NIL

METHOD DrawPointsGraph CLASS GraphPlus

   LOCAL nCategories := 0
   LOCAL nRow, nCol, nData
   LOCAL nAvailableWidth
   LOCAL nTotalBarRatio

   LOCAL nBarWidth, nBarHeight
   LOCAL nCategoryWidth
   LOCAL i, j

   IF hmg_len( ::aData ) > 0
      IF ValType( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      ENDIF
   ENDIF
   IF nCategories == 0
      RETURN NIL
   ENDIF
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth

   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   FOR i := 1 TO hmg_len( ::aData[ 1 ] )
      FOR j := 1 TO hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarHeight := Abs( nData ) * ::nPixelperPoint
         IF nData > 0
            // positive
            BT_DrawFillEllipse ( ::hDC, nRow - nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            IF ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2, nCol, AllTrim( Transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            ENDIF
         ELSE
            // negative
            BT_DrawFillEllipse ( ::hDC, nRow + nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            IF ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2, nCol, if( nData < 0, '-', '' ) + AllTrim( Transform( Abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            ENDIF
         ENDIF
      NEXT j
      nCol += nCategoryWidth
   NEXT i

RETURN NIL

METHOD DrawPieGraph CLASS GraphPlus

   LOCAL fromrow
   LOCAL fromcol
   LOCAL torow
   LOCAL tocol
   LOCAL topleftrow
   LOCAL topleftcol
   LOCAL toprightrow
   LOCAL toprightcol
   LOCAL bottomrightrow
   LOCAL bottomrightcol
   LOCAL bottomleftrow
   LOCAL bottomleftcol
   LOCAL middletopcol
   LOCAL middleleftrow
   LOCAL middlebottomcol
   LOCAL middlerightrow
   LOCAL middlerightcol
   LOCAL fromradialrow
   LOCAL fromradialcol
   LOCAL toradialrow
   LOCAL toradialcol
   LOCAL degrees := {}
   LOCAL cumulative := {}
   LOCAL i, ser_sum := 0, sum, lSum
   LOCAL previos_cumulative
   LOCAL cValue
   LOCAL aSerieValues := {}
   LOCAL nChartSize
   LOCAL aLabelCumulative := {}

   FOR i := 1 TO hmg_len( ::aData[ 1 ] )
      AAdd( aSerieValues, ::aData[ 1, i ] )
   NEXT i
   ::nLeft += 25
   ::nRight -= 25
   ::nBottom -= 25
   ::nTop += 25
   nChartSize := Min( ::nRight - ::nLeft, ::nBottom - ::nTop )

   fromcol := ::nLeft + ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
   tocol := ::nRight - ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
   torow := ::nBottom - ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
   fromrow := ::nTop + ( ( ::nBottom - ::nTop ) - nChartSize ) / 2

   topleftrow := fromrow
   topleftcol := fromcol
   toprightrow := fromrow
   toprightcol := tocol
   bottomrightrow := torow
   bottomrightcol := tocol
   bottomleftrow := torow
   bottomleftcol := fromcol
   middletopcol := fromcol + Int( tocol - fromcol ) / 2
   middleleftrow := fromrow + Int( torow - fromrow ) / 2
   middlebottomcol := fromcol + Int( tocol - fromcol ) / 2
   middlerightrow := fromrow + Int( torow - fromrow ) / 2
   middlerightcol := tocol

   torow := torow + 1
   tocol := tocol + 1
   FOR i := 1 TO HMG_LEN( aSerieValues )
      ser_sum += aSerieValues[ i ]
   NEXT i
   FOR i := 1 TO HMG_LEN( aSerieValues )
      AAdd( degrees, Round( aSerieValues[ i ] / ser_sum * 360, 0 ) )
   NEXT i
   sum := 0
   FOR i := 1 TO HMG_LEN( degrees )
      sum += degrees[ i ]
   NEXT i
   IF sum <> 360
      degrees[ HMG_LEN( degrees ) ] := degrees[ HMG_LEN( degrees ) ] + ( 360 - sum )
   ENDIF

   sum := 0
   lSum := 0
   FOR i := 1 TO HMG_LEN( degrees )
      sum += degrees[ i ]
      AAdd( cumulative, sum )
      lSum += ( degrees[ i ] / 2 )
      AAdd( aLabelCumulative, lSum )
      lSum += ( degrees[ i ] / 2 )
   NEXT i

   previos_cumulative := -1

   fromradialrow := middlerightrow
   fromradialcol := middlerightcol
   FOR i := 1 TO HMG_LEN( cumulative )

      IF cumulative[ i ] == previos_cumulative
         LOOP
      ENDIF
      previos_cumulative := cumulative[ i ]
      DO CASE
      CASE cumulative[ i ] <= 45
         toradialcol := middlerightcol
         toradialrow := middlerightrow - Round( cumulative[ i ] / 45 * ( middlerightrow - toprightrow ), 0 )
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      CASE cumulative[ i ] <= 90 .AND. cumulative[ i ] > 45
         toradialrow := toprightrow
         toradialcol := toprightcol - Round( ( cumulative[ i ] - 45 ) / 45 * ( toprightcol - middletopcol ), 0 )
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      CASE cumulative[ i ] <= 135 .AND. cumulative[ i ] > 90
         toradialrow := topleftrow
         toradialcol := middletopcol - Round( ( cumulative[ i ] - 90 ) / 45 * ( middletopcol - topleftcol ), 0 )
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      CASE cumulative[ i ] <= 180 .AND. cumulative[ i ] > 135
         toradialcol := topleftcol
         toradialrow := topleftrow + Round( ( cumulative[ i ] - 135 ) / 45 * ( middleleftrow - topleftrow ), 0 )
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      CASE cumulative[ i ] <= 225 .AND. cumulative[ i ] > 180
         toradialcol := topleftcol
         toradialrow := middleleftrow + Round( ( cumulative[ i ] - 180 ) / 45 * ( bottomleftrow - middleleftrow ), 0 )
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      CASE cumulative[ i ] <= 270 .AND. cumulative[ i ] > 225
         toradialrow := bottomleftrow
         toradialcol := bottomleftcol + Round( ( cumulative[ i ] - 225 ) / 45 * ( middlebottomcol - bottomleftcol ), 0 )
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      CASE cumulative[ i ] <= 315 .AND. cumulative[ i ] > 270
         toradialrow := bottomleftrow
         toradialcol := middlebottomcol + Round( ( cumulative[ i ] - 270 ) / 45 * ( bottomrightcol - middlebottomcol ), 0 )
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      CASE cumulative[ i ] <= 360 .AND. cumulative[ i ] > 315
         toradialcol := bottomrightcol
         toradialrow := bottomrightrow - Round( ( cumulative[ i ] - 315 ) / 45 * ( bottomrightrow - middlerightrow ), 0 )
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      ENDCASE
   NEXT i
   IF ::lShowValues
      nChartSize := nChartSize / 1.5

      fromcol := ::nLeft + ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
      tocol := ::nRight - ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
      torow := ::nBottom - ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
      fromrow := ::nTop + ( ( ::nBottom - ::nTop ) - nChartSize ) / 2

      topleftrow := fromrow
      topleftcol := fromcol
      toprightrow := fromrow
      toprightcol := tocol
      bottomrightrow := torow
      bottomrightcol := tocol
      bottomleftrow := torow
      bottomleftcol := fromcol
      middletopcol := fromcol + Int( tocol - fromcol ) / 2
      middleleftrow := fromrow + Int( torow - fromrow ) / 2
      middlebottomcol := fromcol + Int( tocol - fromcol ) / 2
      middlerightrow := fromrow + Int( torow - fromrow ) / 2
      middlerightcol := tocol

      previos_cumulative := -1

      FOR i := 1 TO HMG_LEN( cumulative )

         IF aLabelCumulative[ i ] == previos_cumulative
            LOOP
         ENDIF
         cValue := AllTrim( Transform( aSerieValues[ i ], ::cPicture ) ) + ' (' + AllTrim( Transform( aSerieValues[ i ] / ser_sum * 100, '999.99' ) ) + '%)'
         previos_cumulative := aLabelCumulative[ i ]
         DO CASE
         CASE aLabelCumulative[ i ] <= 45
            toradialcol := middlerightcol
            toradialrow := middlerightrow - Round( aLabelCumulative[ i ] / 45 * ( middlerightrow - toprightrow ), 0 )
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
         CASE aLabelCumulative[ i ] <= 90 .AND. aLabelCumulative[ i ] > 45
            toradialrow := toprightrow
            toradialcol := toprightcol - Round( ( aLabelCumulative[ i ] - 45 ) / 45 * ( toprightcol - middletopcol ), 0 )
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
         CASE aLabelCumulative[ i ] <= 135 .AND. aLabelCumulative[ i ] > 90
            toradialrow := topleftrow
            toradialcol := middletopcol - Round( ( aLabelCumulative[ i ] - 90 ) / 45 * ( middletopcol - topleftcol ), 0 )
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
         CASE aLabelCumulative[ i ] <= 180 .AND. aLabelCumulative[ i ] > 135
            toradialcol := topleftcol
            toradialrow := topleftrow + Round( ( aLabelCumulative[ i ] - 135 ) / 45 * ( middleleftrow - topleftrow ), 0 )
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
         CASE aLabelCumulative[ i ] <= 225 .AND. aLabelCumulative[ i ] > 180
            toradialcol := topleftcol
            toradialrow := middleleftrow + Round( ( aLabelCumulative[ i ] - 180 ) / 45 * ( bottomleftrow - middleleftrow ), 0 )
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
         CASE aLabelCumulative[ i ] <= 270 .AND. aLabelCumulative[ i ] > 225
            toradialrow := bottomleftrow
            toradialcol := bottomleftcol + Round( ( aLabelCumulative[ i ] - 225 ) / 45 * ( middlebottomcol - bottomleftcol ), 0 )
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
         CASE aLabelCumulative[ i ] <= 315 .AND. aLabelCumulative[ i ] > 270
            toradialrow := bottomleftrow
            toradialcol := middlebottomcol + Round( ( aLabelCumulative[ i ] - 270 ) / 45 * ( bottomrightcol - middlebottomcol ), 0 )
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
         CASE aLabelCumulative[ i ] <= 360 .AND. aLabelCumulative[ i ] > 315
            toradialcol := bottomrightcol
            toradialrow := bottomrightrow - Round( ( aLabelCumulative[ i ] - 315 ) / 45 * ( bottomrightrow - middlerightrow ), 0 )
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
         ENDCASE
      NEXT i
   ENDIF

RETURN NIL

METHOD DrawBarGraph CLASS GraphPlus

RETURN NIL

METHOD Save( cFileName ) CLASS GraphPlus
   IF PCount() == 0
      cFileName := GetFile ()
   ENDIF
   cFileName := AllTrim( cFileName )
   IF hmg_len( cFileName ) > 0
      IF File( cFileName )
         IF .NOT. MsgYesNo( 'File already Exists! Do you want to overwrite?' )
            RETURN NIL
         ENDIF
      ENDIF
      BT_BitmapSaveFile( ::hBitMap, cFileName, BT_FILEFORMAT_PNG )
   ENDIF

RETURN NIL


METHOD DrawLegends CLASS GraphPlus

   LOCAL nMaxWidth := 0, nMaxHeight := 0
   LOCAL nLegends := hmg_len( ::aLegends )
   LOCAL aSize
   LOCAL nColumns, nRows
   LOCAL nRow
   LOCAL nCol
   LOCAL i, j

   FOR i := 1 TO nLegends
      aSize := BT_DrawTextSize ( ::hDC, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ) )
      nMaxWidth := Max( nMaxWidth, aSize[ 1 ] )
      nMaxHeight := Max( nMaxHeight, aSize[ 2 ] )
   NEXT i
   nMaxWidth := ( nMaxWidth + ::nLegendBoxSize + 10 ) // per legend
   nMaxHeight += 10 // per legend
   ::nLegendWidth := nMaxWidth
   // find columns/rows
   IF ::nLegendPos == 1 // bottom
      nColumns := Int( ( ::nWidth - 20 ) / nMaxWidth )
      IF nColumns == 0 // legend length is not fitting!
         nColumns := 1
      ENDIF
      nRows := Int( nLegends / nColumns ) + if( Mod( nLegends, nColumns ) > 0, 1, 0 )
   ELSE // right
      nColumns := 1
      nRows := nLegends
      IF nLegends * nMaxHeight > ( ::nBottom - ::nTop )
         // all the legends can not fit!
         nRows := Int( ( ::nBottom - ::nTop ) / nMaxHeight )
      ENDIF
   ENDIF
   // draw legends
   IF ::nLegendPos == 1 // bottom
      ::nBottom := ::nBottom - ( nRows * nMaxHeight )
      nRow := ::nBottom + 5
      i := 1
      j := 1
      nCol := ( ::nWidth - nLegends * ::nLegendWidth ) / 2
      DO WHILE i <= nLegends
         IF j > nColumns
            j := 1
            nRow += nMaxHeight
            nCol := ( ::nWidth - nLegends * ::nLegendWidth ) / 2
         ENDIF
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, Max( nMaxHeight - 10, ::nLegendBoxSize ), ::aColors[ i ], ::aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ) )
         nCol += nMaxWidth
         j++
         i++
      ENDDO
   ELSE // right
      ::nRight := ::nRight - nMaxWidth
      nRow := ::nTop + 5
      nCol := ::nRight + 5
      FOR i := 1 TO nRows
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, Max( nMaxHeight - 10, ::nLegendBoxSize ), ::aColors[ i ], ::aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ) )
         nRow += nMaxHeight
      NEXT i
   ENDIF

RETURN NIL

METHOD DrawCategories CLASS GraphPlus

   LOCAL nCategories
   LOCAL nCol
   LOCAL nAvailableWidth
   LOCAL nTotalBarRatio

   LOCAL nBarWidth
   LOCAL nCategoryWidth
   LOCAL i

   nCategories := hmg_len( ::aCategories )
   IF nCategories == 0
      RETURN NIL
   ENDIF
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth

   nCol := ::nLeft + ( nCategoryWidth / 2 )
   FOR i := 1 TO hmg_len( ::aCategories )
      BT_DrawText ( ::hDC, ::nCategoryRow, nCol, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
      nCol += nCategoryWidth
   NEXT i

RETURN NIL

METHOD DrawXYTitles CLASS GraphPlus

   LOCAL aSize
   IF hmg_len( ::cXTitle ) > 0
      aSize := BT_DrawTextSize ( ::hDC, ::cXTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ) )
      BT_DrawText ( ::hDC, ::nBottom - aSize[ 2 ], ( ::nRight - ::nLeft ) / 2, ::cXTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
      ::nBottom := ::nBottom - aSize[ 2 ]
   ENDIF
   IF hmg_len( ::cYTitle ) > 0
      aSize := BT_DrawTextSize ( ::hDC, ::cYTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ) )
      BT_DrawText ( ::hDC, ( ::nBottom - ::nTop ) / 2, ::nLeft, ::cYTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER, BT_TEXT_VERTICAL_ASCENDANT )
      ::nLeft := ::nLeft + aSize[ 2 ]
   ENDIF

RETURN NIL

METHOD InitColors CLASS GraphPlus

   LOCAL i, j, nLegends, nFactor, nCurFactor := 10
   LOCAL aHSL
   LOCAL aColorTheme
   nLegends := hmg_len( ::aLegends )
   IF hmg_len( ::aColors ) = 0 .OR. hmg_len( ::aColors ) < nLegends
      ASize( ::aColors, 0 )
      aColorTheme := ::aColorThemes[ ::nColorTheme ]
      IF nLegends <= hmg_len( aColorTheme )
         FOR i := 1 TO nLegends
            AAdd( ::aColors, aColorTheme[ i ] )
         NEXT i
      ELSE
         FOR i := 1 TO hmg_len( aColorTheme )
            AAdd( ::aColors, aColorTheme[ i ] )
         NEXT i
         i := hmg_len( aColorTheme ) + 1
         j := 1
         nFactor := 10
         DO WHILE i <= hmg_len( ::aLegends )
            IF j > hmg_len( aColorTheme )
               j := 1
               nCurFactor *= nFactor
            ENDIF
            aHSL := GP_RGB2HSL( aColorTheme[ j ] )
            aHSL[ 3 ] := aHSL[ 3 ] + iif( ::lLighter, nCurFactor, -nCurFactor )
            AAdd( ::aColors, GP_HSL2RGB( aHSL ) )
            j++
            i++
         ENDDO
      ENDIF
   ENDIF

RETURN NIL

METHOD SwitchRowsToColumns

   LOCAL aTemp := {}, aLine := {}
   LOCAL cTemp
   LOCAL i
   LOCAL j
   // transpose data
   FOR i := 1 TO hmg_len( ::aData[ 1 ] )
      ASize( aLine, 0 )
      FOR j := 1 TO hmg_len( ::aData )
         AAdd( aLine, ::aData[ j, i ] )
      NEXT j
      AAdd( aTemp, AClone( aLine ) )
   NEXT i
   ASize( ::aData, 0 )
   ::aData := AClone( aTemp )
   // switch legends and categories
   aTemp := AClone( ::aCategories )
   ::aCategories := AClone( ::aLegends )
   ::aLegends := AClone( aTemp )
   // switch xy titles
   cTemp := ::cXTitle
   ::cXTitle := ::cYTitle
   ::cYTitle := cTemp

RETURN NIL

// -------------------------------------------------------------------------- //

FUNCTION findboundary( nNum )

   LOCAL nE, nMax, nMan, nVal, nOffset

   nE := 10
   nVal := 0
   nNum := Abs( nNum )

   DO WHILE .T.

      nMax := 10 ** nE

      IF Int( nNum / nMax ) > 0

         nMan := ( nNum / nMax ) - Int( nNum / nMax )
         nOffset := 1
         nOffset := IF( nMan <= .75, .75, nOffset )
         nOffset := IF( nMan <= .50, .50, nOffset )
         nOffset := IF( nMan <= .25, .25, nOffset )
         nOffset := IF( nMan <= .00, .00, nOffset )
         nVal := ( Int( nNum / nMax ) + nOffset ) * nMax
         EXIT

      ENDIF

      nE--

   ENDDO

RETURN ( nVal )

FUNCTION GetNegativeColor( aRGB )

   LOCAL aNewRGB
   aNewRGB := { 255 - aRGB[ 1 ], 255 - aRGB[ 2 ], 255 - aRGB[ 3 ] }

RETURN aNewRGB

FUNCTION GP_RGB2HSL( aRGB )

   LOCAL nMax, nMin, nDelta
   LOCAL nH, nS, nL, nR, nG, nB

   nR := aRGB[ 1 ] / 255
   nG := aRGB[ 2 ] / 255
   nB := aRGB[ 3 ] / 255
   nMax := Max( nR, Max( nG, nB ) )
   nMin := Min( nR, Min( nG, nB ) )
   nDelta := nMax - nMin

   IF nDelta == 0
      nH := 0
   ELSEIF ( nMax == nR )
      nH := ( ( nG - nB ) / nDelta ) % 6
   ELSEIF ( nMax == nG )
      nH := ( nB - nR ) / nDelta + 2
   ELSE
      nH := ( nR - nG ) / nDelta + 4
   ENDIF
   nH := Int( nH * 60 )
   IF ( nH < 0 )
      nH += 360
   ENDIF
   nL := ( nMax + nMin ) / 2
   nS := iif( nDelta == 0, 0, nDelta / ( 1 - Abs( 2 * nL - 1 ) ) )
   nS := nS * 100
   nL := nL * 100

RETURN { nH, nS, nL }

FUNCTION GP_HSL2RGB( aHSL )

   LOCAL nR := 0, nG := 0, nB := 0, nH, nS, nL, c, x, m
   nH := aHSL[ 1 ]
   nS := aHSL[ 2 ] / 100
   nL := aHSL[ 3 ] / 100
   c := ( 1 - Abs( 2 * nL - 1 ) ) * nS
   x := c * ( 1 - Abs( ( nH / 60 ) % 2 - 1 ) )
   m := nL - c / 2
   IF ( 0 <= nH .AND. nH < 60 )
      nR := c
      nG := x
      nB := 0
   ELSEIF ( 60 <= nH .AND. nH < 120 )
      nR := x
      nG := c
      nB := 0
   ELSEIF ( 120 <= nH .AND. nH < 180 )
      nR := 0
      nG := c
      nB := x
   ELSEIF ( 180 <= nH .AND. nH < 240 )
      nR := 0
      nG := x
      nB := c
   ELSEIF ( 240 <= nH .AND. nH < 300 )
      nR := x
      nG := 0
      nB := c
   ELSEIF ( 300 <= nH .AND. nH < 360 )
      nR := c
      nG := 0
      nB := x
   ENDIF
   nR := Int( ( nR + m ) * 255 )
   nG := Int( ( nG + m ) * 255 )
   nB := Int( ( nB + m ) * 255 )

RETURN { nR, nG, nB }

FUNCTION degreetoradian( nDegree )
RETURN ( nDegree * 0.0174532925 )

FUNCTION radiantodegree( nRadian )
RETURN ( nRadian / 0.0174532925 )
